@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use Class::Tiny;
 
-our $VERSION = '0.000005'; # TRIAL
+our $VERSION = '0.000009'; # TRIAL
 
 # Docs {{{1
 
@@ -20,7 +20,9 @@ C<use Class::Tiny> statement in a package.  This module creates custom
 accessors that behave as standard C<Class::Tiny> accessors except that
 they apply type constraints (C<isa> relationships).  Type constraints
 can come from L<Type::Tiny>, L<MooseX::Types>, L<MooX::Types::MooseLike>,
-L<MouseX::Types>, or L<Specio>.
+L<MouseX::Types>, or L<Specio>.  Alternatively, constraints can be applied
+using the technique described in
+L<"Constraints without a type system"|/CONSTRAINTS WITHOUT A TYPE SYSTEM>.
 
 Example of a class using this package:
 
@@ -29,14 +31,11 @@ Example of a class using this package:
 
     use Type::Tiny;
 
-    my $MediumInteger;
-    BEGIN {
-        # Create the type constraint
-        $MediumInteger = Type::Tiny->new(
-            name => 'MediumInteger',
-            constraint => sub { looks_like_number($_) and $_ >= 10 and $_ < 20 }
-        );
-    }
+    # Create the type constraint
+    use vars::i '$MediumInteger' = Type::Tiny->new(
+        name => 'MediumInteger',
+        constraint => sub { looks_like_number($_) and $_ >= 10 and $_ < 20 }
+    );
 
     use Class::Tiny::ConstrainedAccessor {
         medint => $MediumInteger,           # create accessor sub medint()
@@ -48,29 +47,54 @@ Example of a class using this package:
         med_with_default => 12,
     };
 
+=head1 CONSTRAINTS WITHOUT A TYPE SYSTEM
+
+If you don't want to use L<Type::Tiny> or one of the other type systems listed
+above, you can create your own constraints as two-element arrayrefs.  Example:
+
+    use Class::Tiny::ConstrainedAccessor
+        'field' => [ \&checker_sub, \&message_sub ];
+
+C<checker_sub> and C<message_sub> are used as follows to check C<$value>:
+
+    checker_sub($value) or die get_message($value);
+
+Therefore, C<checker_sub> must return truthy if C<$_[0]> passes the constraint,
+or else falsy.  C<get_message> must return something that can be passed to
+C<die()>, when given a C<$_[0]> that has failed the constraint.
+
+If your profile ever tells you that constraint-checks are on the critical
+path, try custom constraints.  They may give you more control or opportunity
+for optimization than general-purpose type systems.
+
 =head1 SUBROUTINES
 
 =head2 import
 
-Creates the accessors you have requested.  Usage:
+Creates the accessors you have requested.  Basic usage:
 
     use Class::Tiny::ConstrainedAccessor
-        [optional arrayref of option=>value],
         name => constraint
-        [, name => constraint]...  ;
+        [, name => constraint ...]; # ... any number of name=>constraint pairs
 
-The options use an arrayref to make them stand out against the
-C<< name=>constraint >> pairs that come after.
+This also creates a L<BUILD()|Class::Tiny/BUILD> subroutine to check the
+constructor parameters, if a C<BUILD()> doesn't already exist.
 
-This also creates a L<Class::Tiny/BUILD> to check the constructor parameters if
-one doesn't exist.  If one does exist, it creates the same function as
-C<_check_all_constraints> so that you can call it from your own BUILD.  It
-takes the same parameters as C<BUILD>.
+If a C<BUILD()> does exist (e.g., you said C<use subs 'BUILD';>), this package
+will create the same function, taking the same parameters as C<BUILD()> would,
+but call it C<_check_all_constraints()>.   You can call this checker from your
+own C<BUILD()> if you want to.
 
 =head1 OPTIONS
 
-Remember, options are carried in an B<arrayref>.  This is to leave room
-for someday carrying attributes and constraints in a hashref.
+To specify options, pass an B<arrayref> as the first argument on the `use`
+line.  This is to leave room for someday carrying attributes and constraints in
+a hashref.  For example:
+
+    use Class::Tiny::ConstrainedAccessor [ OPTION=>value ],
+        name => constraint ...;
+
+Valid options are:
 
 =over
 
@@ -84,7 +108,7 @@ exists or not.  Example:
     use Class::Tiny::ConstrainedAccessor
         [NOBUILD => 1],
         foo => SomeConstraint;
-
+    # Now $object->_check_all_constraints($args) exists, but not BUILD().
 
 =back
 
@@ -109,10 +133,10 @@ sub import {
         my $constraint = $constraints{$k};
 
         my ($checker, $get_message) =
-                _get_constraint_sub($constraint); # dies on failure
+                _get_constraint_sub($k, $constraint); # dies on failure
 
         my $accessor = _make_accessor($k, $checker, $get_message);
-        $accessors{$k} = [$checker, $get_message];
+        $accessors{$k} = [$checker, $get_message];      # Save for BUILD()
 
         { # Install the accessor
             no strict 'refs';
@@ -134,48 +158,73 @@ sub import {
 } #import()
 
 # _get_constraint_sub: Get the subroutine for a constraint.
+# Takes the constraint name (for debug messages) and the constraint.
 # Returns two coderefs: checker and get_message.
 sub _get_constraint_sub {
-    my ($type) = @_;
+    my ($type_name, $type) = @_;
     my ($checker, $get_message);
 
+    # Handle the custom-constraint format
+    if(ref $type eq 'ARRAY') {
+        die "Custom constraint $type_name must have two elements: checker, get_message"
+            unless scalar @$type == 2;
+        die "$type_name: checker must be a subroutine" unless ref($type->[0]) eq 'CODE';
+        die "$type_name: get_message must be a subroutine" unless ref($type->[1]) eq 'CODE';
+        return @$type;
+    }
+
+    # Get type's name, if any
+    my $name = eval { $type->can('name') || $type->can('description') };
+    $name = $type->$name() if $name;
+
+    # Set default message
+    $name = $type_name unless $name;
+    $get_message = sub { "Value is not a $name" };
+
+    # Try the different types of constraints we know about
     DONE: {
 
-        if ( eval { $type->can('compiled_check') }) { # Type::Tiny
-            $checker = $type->compiled_check();
+        if (ref($type) eq 'CODE') { # MooX::Types::MooseLike
+            $checker = sub { eval { $type->(@_); 1 } };
+                # In profiling, seems to be about on par with `&$type;`.
+            last DONE;
+        }
+
+        die "I don't know how to handle non-reference constraint $type_name"
+            unless ref $type;   # All the rest of the checks use $type->can()
+
+        if ( my $method = eval { $type->can('compiled_check') }) { # Type::Tiny
+            $checker = $type->$method();
             $get_message = sub { $type->get_message($_[0]) };
             last DONE;
         }
 
-        if (my $method = eval { $type->can('inline_check') || $type->can('_inline_check') }) { # Moose
-            $checker = eval { eval sprintf 'sub { my $value = shift; %s }', $type->$method('$value') };
-                # Note: will fail if type cannot be inlined
-            $get_message = sub { 'Constraint failed' };     # TODO
+        if (my $method = eval { $type->can('inline_check') || $type->can('_inline_check') }) { # Specio::Constraint::Simple
+            $checker = eval {
+                my $text = $type->$method('$value');
+                    # $method() will fail if type cannot be inlined.
+                local $SIG{__WARN__} = sub { };     # Eval failure => silent
+                eval "sub { my \$value = shift; $text }"
+            };
             last DONE if $checker;
         }
 
-        if (eval { $type->can('check') } ) { # Moose, Mouse
-            $checker = sub { $type->check(@_) };
-            $get_message = sub { 'Constraint failed' };     # TODO
-            last DONE;
-        }
-
-        if (ref($type) eq 'CODE') { # MooX::Types
-            $checker = sub { eval { $type->(@_); 1 } };
-            $get_message = sub { 'Constraint failed' };     # TODO
+        if (my $method = eval { $type->can('check') } ) { # Moose, Mouse
+            $checker = sub { unshift @_, $type; goto &$method; };
+                # Like $type->check(@_), but profiles as much faster on my
+                # test system.
             last DONE;
         }
 
         if(eval { $type->can('value_is_valid') }) { # Specio::Constraint::Simple
             $checker = sub { $type->value_is_valid(@_) };
-            $get_message = sub { 'Value is not a ' . $type->description };
             last DONE;
         }
 
     } #DONE
 
-    die "I don't know how to use this type (" . (ref($type)||'scalar') . ")"
-        unless $checker and $get_message;
+    die "$type_name: I don't know how to use type " . ref($type)
+        unless $checker;
 
     return ($checker, $get_message);
 } #_get_constraint_sub()
@@ -228,15 +277,15 @@ sub _make_build {
 # Rest of the docs {{{1
 __END__
 
-=head1 AUTHOR
+=head1 AUTHORS
 
-Christopher White, C<< <cxwembedded at gmail.com> >>.  Thanks to
+Created by Christopher White, C<< <cxwembedded at gmail.com> >>.  Thanks to
 Toby Inkster for code contributions.
 
 =head1 BUGS
 
 Please report any bugs or feature requests through the GitHub Issues interface
-at L<https://github.com/cxw42/Class-Tiny-ConstrainedAccessor>.  I will be
+at L<https://github.com/cxw42/Class-Tiny-ConstrainedAccessor/issues>.  I will be
 notified, and then you'll automatically be notified of progress on your bug as
 I make changes.
 
@@ -254,10 +303,6 @@ You can also look for information at:
 
 L<https://github.com/cxw42/Class-Tiny-ConstrainedAccessor>
 
-=item * RT: CPAN's request tracker
-
-L<https://rt.cpan.org/NoAuth/Bugs.html?Dist=Class-Tiny-ConstrainedAccessor>
-
 =item * Search CPAN
 
 L<https://metacpan.org/release/Class-Tiny-ConstrainedAccessor>
